/*  plrt.c    CCMATH mathematics library source code.
 *
 *  Copyright (C)  2000   Daniel A. Atkinson    All rights reserved.
 *  This code may be redistributed under the terms of the GNU library
 *  public license (LGPL). ( See the lgpl.license file for details.)
 * ------------------------------------------------------------------------
 */

/*
 * plrt
 *
 *      Find the real and complex roots of a polynomial with real
 *      coefficients.
 *
 *      typedef struct complex {double re,im;} Cpx;
 *      int plrt(double *cof,int n,Cpx *root,double ra,double rb)
 *        cof = pointer to array containing polynomial coefficients, with
 *              the leading coefficient cof[n] != 0
 *        root = pointer to array of complex structures, containing the
 *               polynomial roots at exit (dimension n)
 *        n = degree of polynomial
 *        ra,rb = initial root estimates, with
 *                 ( rb>0 -> real = ra, imag = rb and
 *                   rb<0 -> real roots ra+rb, ra-rb  )
 *       return: 0 -> normal exit
 *               m>0 -> convergence failure for roots k<m
 *
 *
 *        The polynomial is given by
 *
 *           Pn(x) = Sum(i=0 t0 n) cof[i]*x^i  .
 */

#include <math.h>
#include <stdlib.h>


struct Cpx {double re,im;};


int plrt(double *cof,int n,struct Cpx *root,double ra,double rb)
{ double a,b,s,t,w; int itr,pat; struct Cpx *pr;
  double *cs,*cf,*hf,*p,*q,*ul,test=1.e-28;
  cs=cf=(double *)calloc(2*n,sizeof(double)); hf=cf+n;
  pr=root+n-1; ul=hf+n-1;
  if(rb<0.) rb=ra*ra-rb*rb; else rb=ra*ra+rb*rb; ra*= -2.;
  q=cof+n; s= *q--;
  for(p=cf; p<hf ;) *p++ = *q-- /s;
  for(itr=pat=0;;){
    if(itr==0){
      if(n>2){ a=ra; b=rb;}
      else if(n==2){ a= *cf++; b= *cf;}
      else if(n==1){ pr->re= -(*cf); pr->im=0.; free(cs); return 0;}
     }
    s= -a/2.; t=s*s-b;
    if(t>=0.){ t=sqrt(t); pr->re=s+t; (pr--)->im=0.;
               pr->re=s-t; (pr++)->im=0.; }
    else{ t=sqrt(-t); pr->re=s; (pr--)->im=t;
          pr->re=s; (pr++)->im= -t; }
    if(n==2){ free(cs); return 0;}
    for(p=hf,q=cf; q<hf ;) *p++ = *q++;
    for(p=hf,w=1.; p<ul ;){
      *p++ -=a*w; *p-=b*w; w=*(p-1); }
    t= -(*p--); t+=pr->re* *p; s=pr->im* *p;
    if(t*t+s*s<test){ pr-=2; ul-=2; n-=2; itr=pat=0;
      for(p=cf,q=hf; p<ul ;) *p++ = *q++;
     }
    else if(++itr<30){
           for(p=hf,w=1.; p<ul-2 ;){
             *p++ -=w*a; *p-=w*b; w= *(p-1); }
      t= *p++; q=p+1; s=t*t+w*(b*w-a*t);
      b+=(w*(*p*b- *q*a)+ *q*t)/s; a+=(*p*t- *q*w)/s;
     }
    else{ if(pat==3){ free(cs); return n;}
          itr=0; if(pat++ %2) ra= -ra; else rb= -rb;}
   }
}


/* optional program for testing the above function */

#ifdef USE_MAIN

#include <stdio.h>

int main(int c, char *v[])
{
	if (c < 5) {
		fprintf(stderr, "usage:\n\t%s c0 c1 c2 ... cn ra rb\n", *v);
		//                          0 1  2  3      n+1 c-2 c-1
		return EXIT_FAILURE;
	}
	int n = c-4;
	double cof[n+1];
	for (int i = 0; i <= n; i++)
		cof[i] = atof(v[i+1]);
	double ra = atof(v[c-2]);
	double rb = atof(v[c-1]);

	printf("polynomial of degree %d\n", n);

	struct Cpx root[n];
	int r = plrt(cof, n, root, ra, rb);

	printf("plrt returned %d\nroots =", r);
	for (int i = 0; i < n; i++)
		printf(" (%lf %lf)", root[i].re, root[i].im);
	printf("\n");

	return EXIT_SUCCESS;
}

#endif/*USE_MAIN*/